import time
from json import JSONDecodeError

import msal
from shipyard_templates.exit_code_exception import ExitCodeException
from shipyard_templates.shipyard_logger import ShipyardLogger

logger = ShipyardLogger.get_logger()

FAILED_JOB_STATUSES = {"Failed", "PartialComplete"}
ONGOING_JOB_STATUSES = {"Queued", "InProgress", "Unknown"}
COMPLETE_JOB_STATUSES = {"Completed", "Success"}


def generate_access_token_from_client(bi_instance):
    return msal.ConfidentialClientApplication(
        client_id=bi_instance.client_id,
        client_credential=bi_instance.client_secret,
        authority=f"https://login.microsoftonline.com/{bi_instance.tenant_id}",
    ).acquire_token_for_client(
        scopes=["https://analysis.windows.net/powerbi/api/.default"]
    )


def generate_access_token_from_un_pw(bi_instance):
    return msal.PublicClientApplication(
        client_id=bi_instance.client_id,
        authority=f"https://login.microsoftonline.com/{bi_instance.tenant_id}",
    ).acquire_token_by_username_password(
        username=bi_instance.username,
        password=bi_instance.password,
        scopes=["https://analysis.windows.net/powerbi/api/.default"],
    )


def validate_refresh_object_type(bi_instance, object_type):
    object_type = object_type.lower().strip()
    if object_type not in ["dataset", "dataflow"]:
        raise ValueError
    return object_type


def generate_access_token(bi_instance):
    logger.info("Generating access token")
    try:
        token_data = generate_access_token_from_client(bi_instance)
    except ValueError as e:
        raise ExitCodeException(
            f"Error{e}", bi_instance.EXIT_CODE_INVALID_CREDENTIALS
        ) from e

    except Exception as e:
        raise ExitCodeException(
            f"An unexpected error occurred while generating access token. Error{e}",
            bi_instance.EXIT_CODE_UNKNOWN_ERROR,
        ) from e
    else:
        if token := token_data.get("access_token"):
            logger.info("Access token generated successfully")
            return token
        if "error" in token_data:
            error_type = token_data.get("error")
            error_description = token_data.get("error_description")
            if error_type in {"invalid_grant", "invalid_client"}:
                raise ExitCodeException(
                    error_description, bi_instance.EXIT_CODE_INVALID_INPUT
                )
            else:
                raise ExitCodeException(
                    f"Error TYPE: {error_type}, Error: {error_description}",
                    bi_instance.EXIT_CODE_UNKNOWN_ERROR,
                )


def handle_error_response(bi_instance, response):
    try:
        logger.debug(response.text)
        response_details = response.json()
        error_info = response_details.get("error", {})

    except JSONDecodeError:
        logger.warning("Response was not JSON...handling as text")
        error_info = {"message": response.text} if response.text else {}

    if error_info.get("code") == "DailyDataflowRefreshLimitExceeded":
        raise ExitCodeException(
            error_info.get(
                "message",
                f"<Daily Dataflow Refresh Limit Exceeded> \n {response.text}",
            ),
            bi_instance.EXIT_CODE_RATE_LIMIT,
        )
    if response.status_code == 401:
        raise ExitCodeException(
            error_info.get("message", f"<Unauthorized> \n {response.text}"),
            bi_instance.ERROR_CODE_INVALID_CREDENTIALS,
        )
    elif response.status_code == 403:
        raise ExitCodeException(
            error_info.get("message", f"<Forbidden> \n {response.text}"),
            bi_instance.ERROR_CODE_INVALID_CREDENTIALS,
        )
    elif response.status_code == 404:
        raise ExitCodeException(
            error_info.get("message", f"<Not Found\n {response.text}"),
            bi_instance.EXIT_CODE_INVALID_INPUT,
        )
    elif response.status_code == 400:
        raise ExitCodeException(
            error_info.get("message", f"<Bad Request\n {response.text}"),
            bi_instance.EXIT_CODE_BAD_REQUEST,
        )
    elif response.status_code == 405:
        raise ExitCodeException(
            error_info.get("message", f"<Method Not Allowed\n {response.text}"),
            bi_instance.EXIT_CODE_BAD_REQUEST,
        )
    elif response.status_code == 429:
        raise ExitCodeException(
            error_info.get("message", f"<Too Many Requests\n {response.text}"),
            bi_instance.EXIT_CODE_RATE_LIMIT,
        )
    else:
        raise ExitCodeException(
            f"Unknown Error<{response.status_code}>: {response.text}",
            bi_instance.EXIT_CODE_UNKNOWN_ERROR,
        )


def wait_for_dataset_refresh_completion(
        bi_instance, group_id, dataset_id, request_id, wait_time=60
):
    logger.info("Waiting for refresh to complete")
    job_status = "Unknown"

    while job_status in ONGOING_JOB_STATUSES:
        job_status = bi_instance.check_recent_dataset_refresh_by_request_id(
            group_id, dataset_id, request_id
        ).get("status")

        if job_status in COMPLETE_JOB_STATUSES:
            logger.info(f"Job completed with status {job_status}")
            logger.info("Refresh completed")
            return

        if job_status in FAILED_JOB_STATUSES:
            raise ExitCodeException(
                f"Dataset refresh failed with status {job_status}",
                bi_instance.EXIT_CODE_FAILED_REFRESH_JOB,
            )

        elif job_status in ONGOING_JOB_STATUSES:
            logger.info(
                f"Job currently in {job_status}. Waiting {wait_time} seconds before checking again."
            )
            time.sleep(wait_time)
        else:
            raise ExitCodeException(
                f"Unknown job status {job_status}",
                bi_instance.EXIT_CODE_UNKNOWN_REFRESH_JOB_STATUS,
            )


def wait_for_dataflow_refresh_completion(
        bi_instance, group_id, dataflow_id, wait_time=60
):
    logger.info("Waiting for refresh to complete")
    job_status = "Unknown"

    while job_status in ONGOING_JOB_STATUSES:
        transactions = bi_instance.get_dataflow_transactions(group_id, dataflow_id)
        sorted_transactions = sorted(
            transactions["value"], key=lambda x: x["startTime"], reverse=True
        )

        if len(sorted_transactions) == 0:
            raise ExitCodeException(
                f"No transactions found for dataflow {dataflow_id}",
                bi_instance.EXIT_CODE_INVALID_INPUT,
            )

        job_status = sorted_transactions[0].get(
            "status"
        )  # Get the latest transaction status

        if job_status in COMPLETE_JOB_STATUSES:
            logger.info(f"Job completed with status {job_status}")
            logger.info("Refresh completed")
            return

        if job_status in FAILED_JOB_STATUSES:
            raise ExitCodeException(
                f"Dataflow refresh failed with status {job_status}",
                bi_instance.EXIT_CODE_FAILED_REFRESH_JOB,
            )

        elif job_status in ONGOING_JOB_STATUSES:
            logger.info(
                f"Job currently in {job_status}. Waiting {wait_time} seconds before checking again."
            )
            time.sleep(wait_time)
        else:
            raise ExitCodeException(
                f"Unknown job status {job_status}",
                bi_instance.EXIT_CODE_UNKNOWN_REFRESH_JOB_STATUS,
            )



import json

from requests import request
from shipyard_templates import DataVisualization, ExitCodeException, ShipyardLogger

from shipyard_microsoft_power_bi import microsoft_power_bi_utils as bi_utils

logger = ShipyardLogger.get_logger()


class MicrosoftPowerBiClient(DataVisualization):
    """
    A client for interacting with Microsoft Power BI.

    This class provides methods to authenticate, refresh datasets, manage dataflows, and retrieve data.

    Attributes:
        BASE_URL (str): The base URL for the Power BI API.
        EXIT_CODE_FAILED_REFRESH_JOB (int): Exit code for failed refresh job.
        EXIT_CODE_DATAFLOW_REFRESH_ALREADY_IN_PROGRESS (int): Exit code for when a dataflow refresh is already in progress.
    """

    BASE_URL = "https://api.powerbi.com/v1.0/myorg"
    EXIT_CODE_FAILED_REFRESH_JOB = 101
    EXIT_CODE_DATAFLOW_REFRESH_ALREADY_IN_PROGRESS = 102
    EXIT_CODE_UNKNOWN_REFRESH_JOB_STATUS = 103

    def __init__(self, client_id: str, client_secret: str, tenant_id: str, **kwargs):
        """
        Initializes the MicrosoftPowerBiClient with the provided credentials.

        Args:
            client_id (str): Client ID for Power BI API authentication.
            client_secret (str): Client Secret for Power BI API authentication.
            tenant_id (str): Tenant ID for Power BI API authentication.
            **kwargs: Additional keyword arguments passed to the superclass.
        """

        super().__init__(**kwargs)
        self.access_token = None
        self.client_id = client_id
        self.client_secret = client_secret
        self.tenant_id = tenant_id

    def _request(self, endpoint: str, method: str = "GET", **kwargs):
        """
        A helper function to make a request to the Power BI API.

        @param endpoint: The endpoint to make the request to.
        @param method: The HTTP method to use.
        @param kwargs: Additional keyword arguments to pass to the request.
        @return: The response from the request.
        """
        if self.access_token is None:
            self.access_token = bi_utils.generate_access_token(self)

        logger.debug(f"Attempting to {method} {endpoint}")

        response = request(
            method=method,
            url=endpoint,
            headers={
                "Authorization": f"Bearer {self.access_token}",
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
            **kwargs,
        )
        logger.debug(f"Response Code: {response.status_code}")
        logger.debug(response.text)

        if response.ok:
            if response.status_code == 200:
                try:
                    return response.json()
                except json.decoder.JSONDecodeError:
                    return response.text

            if response.status_code == 202:
                logger.debug("No Content")
                return response
        else:
            bi_utils.handle_error_response(self, response)

    def connect(self):
        """
        Confirms that the client can connect to the Power BI API.

        @return: 1 if the connection fails, 0 otherwise.
        """
        try:
            bi_utils.generate_access_token(self)
        except ExitCodeException as e:
            logger.authtest(e)
            return 1
        else:
            return 0

    def refresh(
            self,
            object_type: str,
            group_id: str,
            object_id: str,
            wait_for_completion: bool,
            wait_time: int = 60,
    ):
        """
        Triggers a refresh job for the specified object with the option to wait that job to complete.

        @param object_type: Either "dataset" or "dataflow".
        @param group_id: The ID of the group/workspace that the object belongs to.
        @param object_id: Either the ID of the dataset or dataflow to refresh.
        @param wait_for_completion: If True, waits for the refresh job to complete.
        @param wait_time: Used if wait_for_completion is True.The number of seconds to wait for the refresh job to complete.
        @return: None
        """
        if object_type == "dataset":
            self.refresh_dataset(group_id, object_id, wait_for_completion, wait_time)
        elif object_type == "dataflow":
            self.refresh_dataflow(group_id, object_id, wait_for_completion, wait_time)
        else:
            raise ExitCodeException(
                f"Refresh object type {object_type} not implemented",
                self.EXIT_CODE_INVALID_INPUT,
            )

    def refresh_dataset(
            self,
            group_id: str,
            dataset_id: str,
            wait_for_completion: bool,
            wait_time: int = 60,
    ):
        """
        Triggers a refresh job for the specified dataset with the option to wait that job to complete.
        resource: https://docs.microsoft.com/en-us/rest/api/power-bi/datasets/refreshdataset
        @param group_id: Group/Workspace ID
        @param dataset_id: Dataset ID
        @param wait_for_completion: If True, waits for the refresh job to complete.
        @param wait_time: if wait_for_completion is True, the number of seconds to wait for the refresh job to complete.
        @return: response from the request or None if wait_for_completion is True.
        """
        logger.info("Triggering dataset refresh...")
        response = self._request(
            f"{self.BASE_URL}/groups/{group_id}/datasets/{dataset_id}/refreshes",
            method="POST",
        )
        logger.info("Dataset refresh triggered")
        if wait_for_completion:
            request_id = response.headers.get("RequestId")
            bi_utils.wait_for_dataset_refresh_completion(
                self, group_id, dataset_id, request_id, wait_time=wait_time
            )
        else:
            return response

    def refresh_dataflow(
            self,
            group_id: str,
            dataflow_id: str,
            wait_for_completion: bool,
            wait_time: int = 60,
    ):
        """
        Triggers a refresh job for the specified dataflow with the option to wait that job to complete.
        resource: https://docs.microsoft.com/en-us/rest/api/power-bi/dataflows/refreshdataflow

        @param group_id: Group/Workspace ID
        @param dataflow_id: Dataflow ID
        @param wait_for_completion: If True, waits for the refresh job to complete.
        @param wait_time: If wait_for_completion is True, the number of seconds to wait for the refresh job to complete.
        @return: response from the request or None if wait_for_completion is True.
        """

        data = {"refreshRequest": "ShipyardRefresh"}
        logger.info("Triggering dataflow refresh...")
        response = self._request(
            f"{self.BASE_URL}/groups/{group_id}/dataflows/{dataflow_id}/refreshes",
            method="POST",
            data=json.dumps(data),
        )
        if wait_for_completion:
            bi_utils.wait_for_dataflow_refresh_completion(
                self, group_id, dataflow_id, wait_time=wait_time
            )
        else:
            logger.info("Dataflow refresh triggered")
            return response

    def get_dataflow_transactions(self, group_id: str, dataflow_id: str):
        """
        Gets the transactions for the specified dataflow.
        resource: https://docs.microsoft.com/en-us/rest/api/power-bi/dataflows/getdataflowtransactions

        @param group_id: Group/Workspace ID
        @param dataflow_id: Dataflow ID
        @return: response from the request
        """
        return self._request(
            f"{self.BASE_URL}/groups/{group_id}/dataflows/{dataflow_id}/transactions",
            method="GET",
        )

    def check_recent_dataset_refresh_by_request_id(
            self,
            group_id: str,
            dataset_id: str,
            request_id: str,
            number_of_refreshes: int = 10,
    ):
        """
        Gets the status of the most recent refreshes for the specified dataset and checks if the specified request ID is
        present in those refreshes.
        resource: https://docs.microsoft.com/en-us/rest/api/power-bi/datasets/getdatasetrefreshes

        @param group_id: Group/Workspace ID
        @param dataset_id: Dataset ID
        @param request_id: Request ID to check for. This is the ID returned by the refresh request.
        @param number_of_refreshes: The number of recent refreshes to check.
        @return: Response from the request.
        """
        response = self._request(
            f"{self.BASE_URL}/groups/{group_id}/datasets/{dataset_id}/refreshes?$top={number_of_refreshes}",
            method="GET",
        )
        jobs = response.get("value")
        for job in jobs:
            if job.get("requestId") == request_id:
                current_status = job.get("status")
                logger.debug(f"Current status: {current_status}")
                return job
        raise ExitCodeException(
            f"Request ID {request_id} not found within the last {number_of_refreshes} refreshes",
            self.EXIT_CODE_INVALID_INPUT,
        )

    def check_dataset_refresh_by_refresh_id(
            self, group_id: str, dataset_id: str, refresh_id: str
    ):
        """
        Gets the status of the specified refresh.
        resource: https://docs.microsoft.com/en-us/rest/api/power-bi/datasets/getdatasetrefreshbyid

        @param group_id: Group/Workspace ID
        @param dataset_id: Dataset ID
        @param refresh_id: Refresh ID
        @return: response from the request
        """
        return self._request(
            f"{self.BASE_URL}/groups/{group_id}/datasets/{dataset_id}/refreshes/{refresh_id}",
            method="GET",
        )

    def get_dataset(self, group_id: str, dataset_id: str):
        """
        Gets the specified dataset.
        resource: https://docs.microsoft.com/en-us/rest/api/power-bi/datasets/getdatasetbyid

        @param group_id: Group/Workspace ID
        @param dataset_id: Dataset ID
        @return: Response from the request
        """
        return self._request(
            f"{self.BASE_URL}/groups/{group_id}/datasets/{dataset_id}", method="GET"
        )

    def get_datasets(self, group_id: str):
        """
        Gets the datasets in the specified group.
        resource: https://docs.microsoft.com/en-us/rest/api/power-bi/datasets/getdatasetsingroup

        @param group_id: Group/Workspace ID
        @return: Response from the request
        """
        return self._request(
            f"{self.BASE_URL}/groups/{group_id}/datasets", method="GET"
        )

    def get_dataflows(self, group_id: str):
        """
        Gets the dataflows in the specified group.
        resource: https://docs.microsoft.com/en-us/rest/api/power-bi/dataflows/getdataflows

        @param group_id: Group/Workspace ID
        @return: Response from the request
        """
        return self._request(
            f"{self.BASE_URL}/groups/{group_id}/dataflows", method="GET"
        )



import os
import sys

from shipyard_microsoft_power_bi import MicrosoftPowerBiClient
from shipyard_templates.shipyard_logger import ShipyardLogger

logger = ShipyardLogger.get_logger()
logger.setLevel("AUTHTEST")


def main():
    sys.exit(
        MicrosoftPowerBiClient(
            client_id=os.getenv("MICROSOFT_POWER_BI_CLIENT_ID"),
            client_secret=os.getenv("MICROSOFT_POWER_BI_CLIENT_SECRET"),
            tenant_id=os.getenv("MICROSOFT_POWER_BI_TENANT_ID"),
        ).connect()
    )


if __name__ == "__main__":
    main()



import sys
import argparse
from shipyard_microsoft_power_bi import MicrosoftPowerBiClient
from shipyard_templates.exit_code_exception import ExitCodeException


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--client-id", dest="client_id", required=True)
    parser.add_argument("--client-secret", dest="client_secret", required=True)
    parser.add_argument("--tenant-id", dest="tenant_id", required=True)
    parser.add_argument("--group-id", dest="group_id", required=True)
    parser.add_argument("--object-type", dest="object_type", required=True)
    parser.add_argument("--object-id", dest="object_id", required=True)
    parser.add_argument(
        "--wait-for-completion", dest="wait_for_completion", required=True
    )
    parser.add_argument("--poke-interval", dest="poke_interval", required=False)

    return parser.parse_args()


def validate_args(args):
    """
    Handle datatype conversions and validations for arguments

    @param args: Namespace object containing arguments
    @return: Tuple containing wait_for_completion and wait_time
    """

    if args.poke_interval is None or args.poke_interval == "":
        wait_time = 1
    else:
        wait_time = int(args.poke_interval) * 60

    if wait_time < 0:
        raise ExitCodeException(
            "Error: poke-interval must be greater than 0",
            MicrosoftPowerBiClient.EXIT_CODE_INVALID_INPUT,
        )
    elif wait_time > 60:
        raise ExitCodeException(
            "Error: poke-interval must be less than or equal to 60",
            MicrosoftPowerBiClient.EXIT_CODE_INVALID_INPUT,
        )

    wait_for_completion = None
    if type(args.wait_for_completion) is str:
        if args.wait_for_completion.upper() == "TRUE":
            wait_for_completion = True
        elif args.wait_for_completion.upper() == "FALSE":
            wait_for_completion = False

    elif type(args.wait_for_completion) is bool:
        wait_for_completion = args.wait_for_completion

    return wait_for_completion, wait_time


def main():
    try:  # Initialize client to ensure client.logger is available
        args = get_args()

        client = MicrosoftPowerBiClient(
            client_id=args.client_id,
            client_secret=args.client_secret,
            tenant_id=args.tenant_id,
        )
    except ExitCodeException as e:
        print(e)
        sys.exit(e.exit_code)
    except Exception as e:
        print(f"Unexpected error: {e}")
        sys.exit(MicrosoftPowerBiClient.EXIT_CODE_UNKNOWN_ERROR)

    try:
        wait_for_completion, wait_time = validate_args(args)

        client.refresh(
            object_type=args.object_type,
            object_id=args.object_id,
            group_id=args.group_id,
            wait_for_completion=wait_for_completion,
            wait_time=wait_time,
        )
    except ExitCodeException as e:
        client.logger.error(e)
        sys.exit(e.exit_code)
    except Exception as e:
        client.logger.error(f"Unexpected error: {e}")
        sys.exit(client.EXIT_CODE_UNKNOWN_ERROR)


if __name__ == "__main__":
    main()
